/** @file
 *  @brief Map string to wdf normalisation code
 */
/* Warning: This file is generated by ./weight/collate-wdf-norm - do not modify directly! */
/* Copyright (C) 2020 Dipanshu Garg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef XAPIAN_INCLUDED_WDF_NORM_DISPATCH_H
#define XAPIAN_INCLUDED_WDF_NORM_DISPATCH_H

static const unsigned char wdf_norm_tab[] = {
    11,

      1,   1,  19,   8,   1,   0,  32,   1,
      1,   1,  57,

    (1 - 1),
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::SQUARE), 'S','Q','U','A','R','E',

    (2 - 1),
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::NONE), 'N','O','N','E',
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::SQRT), 'S','Q','R','T',

    (3 - 1),
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::AUG), 'A','U','G',
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::LOG), 'L','O','G',
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::MAX), 'M','A','X',

    (3 - 1),
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::AUG_LOG), 'A','U','G','_','L','O','G',
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::BOOLEAN), 'B','O','O','L','E','A','N',
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::PIVOTED), 'P','I','V','O','T','E','D',

    (2 - 1),
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::AUG_AVERAGE), 'A','U','G','_','A','V','E','R','A','G','E',
    static_cast<unsigned char>(Xapian::TfIdfWeight::wdf_norm::LOG_AVERAGE), 'L','O','G','_','A','V','E','R','A','G','E'
};

#endif
